//
// 64drive USB loader
// (c) 2011 Retroactive
// 
// This code is open-source. Feel free to improve it but know that if you screw up
// the firmware update you'll need a JTAG cable to re-configure it.
//
// You're welcome to port to other platforms. Send me a link or zip file.
// Uses FTDI D2XX drivers. http://www.ftdichip.com/Drivers/D2XX.htm
//
// http://64drive.retroactive.be
//
// 0.1 partial libftdi port by hcs
//

#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <unistd.h>
#include <ftdi.h>

#define CHUNK_SIZE (512*1024)
unsigned char buffer[CHUNK_SIZE];

int		save_type = 0;
int		save_num = 4;
const char	*save_type_list[] = {"None (default)","EEPROM 4k", "EEPROM 16k", "SRAM 256kbit", "FlashRAM 1Mbit", "FlashRAM 1Mbit (for Pokemon Stadium 2)", "SRAM 768Kbit (for Dezaemon 3D)"};
const char	*compile_date = __TIMESTAMP__;
long	sdram_addr = 0;

unsigned int file_size = 0;
char image_filename[2048];


void fail(int st);

inline int SwapEndian(int val)
{
	return ((val<<24) ) | 
		   ((val<<8)  & 0x00ff0000) |
		   ((val>>8)  & 0x0000ff00) | 
		   ((val>>24) );
}

void write_header_bus(unsigned char* tx_buffer, int address, unsigned short data)
{
	tx_buffer[0] = 0x90; // operation
	tx_buffer[1] = 0x08; // length of packet

	*(int *)&tx_buffer[2] = SwapEndian(address);
	tx_buffer[6] = data & 0xff;
	tx_buffer[7] = (data >> 8) & 0xff;
}

void write_header_bus_long(unsigned char* tx_buffer, int address, unsigned short data)
{
	tx_buffer[0] = 0x91; // operation
	tx_buffer[1] = 0x08; // length of packet

	*(int *)&tx_buffer[2] = SwapEndian(address);
	tx_buffer[6] = data & 0xff;
	tx_buffer[7] = (data >> 8) & 0xff;
}

void write_header(unsigned char* tx_buffer, unsigned char operation, int sdram_start, int num_words)
{
	memset(tx_buffer, 0, 32);
	tx_buffer[0] = operation; // operation
	tx_buffer[1] = 0x0A; // length of packet

	*(int *)&tx_buffer[2] = SwapEndian(num_words);
	*(int *)&tx_buffer[6] = SwapEndian(sdram_start);
}


#define MODE_UNSPECIFIED	0
#define MODE_LOADIMAGE		1
#define MODE_SETSAVE		2

#define MODE_BOOTLOADER		16
#define MODE_FPGACONFIG		32

#define ERR_TIMEOUT			9
#define ERR_CANTOPENDEVICE	10
#define ERR_CANTOPENIMAGE	11
#define ERR_INVALIDSAVE		12
#define ERR_INVALIDBTLDR	13
#define ERR_INVALIDFIRM		14

int main(int argc, char ** argv)
{
	int i;
	int last_option;
	int mode = MODE_UNSPECIFIED;
	uint32_t rx_queue;
	uint32_t tx_queue;
	uint32_t event_status;
	uint32_t ft_devices;
	int status;
    struct ftdi_context *ftdi = NULL;
    struct ftdi_device_list *devInfo = NULL;
	int found = 0;
    struct usb_device *found_dev = NULL;

	uint32_t bytes_written;
	uint32_t bytes_read;
	FILE* fp;

	printf("\n");
	printf(" 64drive USB Loader\n----------------------\n");
	printf(" (c) 2011 Retroactive\n");
	printf(" Compiled %s\n", compile_date);
	printf(" In case of transfer lock-up, press Ctrl-C to cancel before unplugging\n\n");
	for (i = 1; i < argc; ++i)  {
		if (argv[i][0] == '-'){ // looks like an argument 
       
			switch(argv[i][1]){ // look at next char
            
                default: 
                    break;
            }
            last_option = argv[i][1];
        }else{
            // this is not a -option argument, so it must be a textual argument        
            switch(last_option) 
            {   
				case 'a':
					sscanf(argv[i], "%lx", &sdram_addr);
					break;
                case 'l': 
                    strcpy(image_filename, argv[i]);
					mode |= MODE_LOADIMAGE;
                    break;
				case 's':
					sscanf(argv[i], "%d", &save_type);
					mode |= MODE_SETSAVE;
					break;
#if 0
                case 'b': 
                    strcpy(image_filename, argv[i]);
					mode = MODE_BOOTLOADER;
					sdram_addr = 0;
                    break;
                case 'f':
					strcpy(image_filename, argv[i]);
					mode = MODE_FPGACONFIG;
					sdram_addr = 0;
					break;
#endif
				
				
                default: 
                    break;
            }
            last_option = '\0';
        }
    }

	if(mode == MODE_UNSPECIFIED){
        // filename was not given, complain
        printf(" - Invalid parameter(s)\n - Parameters:\n\n");
		printf("\t-l <filename.z64>\t Load image to offset in SDRAM\n");
		printf("\t-a <offset in hex>\t Set offset (default is 0x0)\n");
		printf("\t-s <save type>\t\t Set save type \n");
		for(i = 0; i < save_num+1; i++){
			printf("\t\t\t\t   %d - %s\n", i, save_type_list[i]);
		}

#if 0
		printf(" - Update commands\n\n");
		printf("\t-b <filename.bin>\t Update bootloader\n");
		printf("\t-f <filename.bin>\t Update firmware (danger of bricking)\n");
#endif
        exit(1);
    }
	
	
	////////////////////////////////////////////////////
	// initialize library
    //
    ftdi = ftdi_new();
    if (!ftdi)
    {
		printf(" *** Couldn't create ftdi_context.\n");
        exit(1);
    }

	////////////////////////////////////////////////////
	// enumerate
	//
    {
        status = ftdi_usb_find_all(ftdi, &devInfo, 0x0403, 0x6010);
		if(status > 0) {
            ft_devices = status;
            printf("Found %d devices\n", ft_devices);
			for(int i = 0; i < ft_devices; i++){
                char manufacturer[256];
                char description[256];
                char serial[256];

                status = ftdi_usb_get_strings(ftdi, devInfo[i].dev,
                    manufacturer, sizeof(manufacturer),
                    description, sizeof(description),
                    serial, sizeof(serial));

                if (status != 0)
                {
                    printf("ftdi_usb_get_strings returned status %d\n", status);
                    exit(1);
                }

				printf("    [Device %d]\n", i);
				//printf("     -- Flags\t\t0x%x\n",devInfo[i].Flags); 
				//printf("     -- Type\t\t0x%x\n",devInfo[i].Type); 
				//printf("     -- ID\t\t0x%x\n",devInfo[i].ID); 
				//printf("     -- LocId\t\t0x%x\n",devInfo[i].LocId); 
                printf("     -- Manufacturer\t%s\n",manufacturer);
				printf("     -- SerialNumber\t%s\n",serial); 
				printf("     -- Description\t\"%s\"\n",description); 

                if (!strcmp(description, "64drive USB device"))
                {
                    found = 1;
                    found_dev = devInfo[i].dev;
                }
			}
		}
        else
        {
            printf(" *** Failed to enumerate or find devices (status %d)\n", status);
            exit(1);
        }

	}

	if(!found){
		printf(" - No attached devices are 64drives\n");
		exit(-1);
	}

	////////////////////////////////////////////////////
	// open device
	//
	//printf(" - Opening device... ");
    status = ftdi_usb_open_dev(ftdi, found_dev);
	if(status == 0){
		//printf("OK!\n");
	}else{
		printf(" *** Couldn't open device, status %d.\n", status);
		fail(ERR_CANTOPENDEVICE);
	}

#if 0
	// set read/write timeouts
	fail(FT_SetTimeouts(handle, 10000, 10000));
	fail(FT_ResetDevice(handle)); // reset is done implicitly on open
#endif


	////////////////////////////////////////////////////
	// open image file
	//
	char *filename_start = strrchr(image_filename, '\\');
	if(filename_start == NULL) filename_start = image_filename-1;
	printf(" - Opening file %s ", filename_start+1);
	fp = fopen(image_filename, "rb");
	if(fp == NULL) {
		printf("\n\n *** Couldn't open file\n");
		fail(ERR_CANTOPENIMAGE);
	}

	fseek(fp, 0, SEEK_END);
	file_size = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	printf("(%d bytes)\n", file_size);
	
	unsigned char tx_buffer[64];
	unsigned char rx_buffer[16];


	/////////////////////////////////////////////////////
	// set save type
	//
	if(mode & MODE_SETSAVE){
		if(save_type < 0 || save_type > save_num){
			printf("\n *** Invalid save specified (%d)\n", save_type);
			fail(ERR_INVALIDSAVE);
		}
		printf(" - Setting save type to %s\n", save_type_list[save_type]);
		memset(tx_buffer, 0, 32);
		tx_buffer[0] = 0xd0;	// operation
		tx_buffer[1] = 4;		// length of packet
		tx_buffer[2] = save_type;
        status = ftdi_write_data(ftdi, tx_buffer, 4);
        if (status > 0)
        {
            bytes_written = status;
        }
        else
        {
            printf(" *** Failed writing save type\n");
            exit(1);
        }
	}
	

#if 0
	/////////////////////////////////////////////////////
	// bulk erase bootloader
	//
	if(mode & MODE_BOOTLOADER){
		fread(tx_buffer, 16, 1, fp);
		if(strncmp(tx_buffer, "RETROACTIVE_2011", 16) != 0){
			printf(" * Invalid header. Make sure you have the right file!\n");
			fail(ERR_INVALIDBTLDR);
		}
		printf("\n\n");
		printf(" * Warning: about to erase and rewrite your bootloader!!!\n");
		printf(" * Be sure you have a valid file.\n");
		printf(" * Continue? (y/n) ");
		if(getchar() != 'y') fail(MODE_BOOTLOADER);
		printf("\n * Please ensure your cartridge is not in the N64!\n");
		printf(" * Press enter to continue.\n");
		getchar();getchar();

		if(file_size != 524288+16){
			printf(" * Bootloader image is not the proper size\n");
			fail(MODE_BOOTLOADER);
		}

		printf(" - Bulk erasing bootloader\n");
			
		memset(tx_buffer, 0, 32);

		// switch cpld to whole-chip access
		write_header_bus(tx_buffer, 0x17000000, 0x0000);
		fail(FT_Write(handle, tx_buffer, 8, &bytes_written));	

		write_header_bus(tx_buffer, 0x5555*2, 0xAA00);
		fail(FT_Write(handle, tx_buffer, 8, &bytes_written));
		write_header_bus(tx_buffer, 0x2AAA*2, 0x5500);
		fail(FT_Write(handle, tx_buffer, 8, &bytes_written));
		write_header_bus(tx_buffer, 0x5555*2, 0x8000);
		fail(FT_Write(handle, tx_buffer, 8, &bytes_written));

		write_header_bus(tx_buffer, 0x5555*2, 0xAA00);
		fail(FT_Write(handle, tx_buffer, 8, &bytes_written));
		write_header_bus(tx_buffer, 0x2AAA*2, 0x5500);
		fail(FT_Write(handle, tx_buffer, 8, &bytes_written));
		write_header_bus(tx_buffer, 0x5555*2, 0x1000);
		fail(FT_Write(handle, tx_buffer, 8, &bytes_written));
		sleep(1);
	}
	

	
	/////////////////////////////////////////////////////
	// write bootloader
	//
	if(mode & MODE_BOOTLOADER){
		file_size = 524288;
		int bytes_left = file_size;
		int bytes_done = 0;
		int bytes_do;
		unsigned short word;

		sdram_addr = 0;
		printf(" - Writing bootloader\n");
		printf(" - You might want to make some coffee\n");
		while(1){
			if(bytes_left >= 2) 
				bytes_do = 2;
			else
				bytes_do = bytes_left;
			if(bytes_do <= 0) break;


			fread(&word, 2, 1, fp);
			memset(tx_buffer, 0, 32);
			write_header_bus(tx_buffer,			0x5555*2, 0xAA00);
			write_header_bus(&tx_buffer[8],		0x2AAA*2, 0x5500);
			write_header_bus(&tx_buffer[16],	0x5555*2, 0xA000);
			write_header_bus_long(&tx_buffer[24], sdram_addr, word);
			fail(FT_Write(handle, tx_buffer, 32, &bytes_written));

			bytes_left -= bytes_do;
			bytes_done += bytes_do;
			sdram_addr += bytes_do;

			// progress bar
			int blocks_done = ((float)bytes_done / file_size)*64;
			int blocks_left = 64-blocks_done;
			printf("\r   ");
			for(int i = 0; i < blocks_done; i++) printf("\xdb");
			for(int i = 0; i < blocks_left; i++) printf("\xb0");
		}
		printf("\n - Success!\n");
		printf(" - Please unplug the USB cable and plug it back in.\n");
	}
	
	
	/////////////////////////////////////////////////////
	// read bootloader
	//
	/*
	printf(" - Reading bootloader\n");
	FILE* wp = fopen("dump.bin", "wb");


	unsigned short word;
	while(1){

		memset(tx_buffer, 0, 32);
		tx_buffer[0] = 0xA1; // operation
		tx_buffer[1] = 0x08; // length of packet
		*(int *)&tx_buffer[2] = SwapEndian(sdram_addr);
		fail(FT_Write(handle, tx_buffer, 6, &bytes_written));
		fail(FT_Read(handle, rx_buffer, 2, &bytes_read));

		fwrite(rx_buffer, 2, 1, wp);

		sdram_addr += 2;
		if(sdram_addr >= 4096) break;

		if(sdram_addr % 256 == 0) fflush(wp);
	}

	fclose(wp);
	*/
	
	/////////////////////////////////////////////////////
	// bulk erase config
	//
	
	if(mode & MODE_FPGACONFIG){
		fread(tx_buffer, 16, 1, fp);
		if(strncmp(tx_buffer, "RETROACTIVE-2011", 16) != 0){
			printf(" * Invalid header. Make sure you have the right file!\n");
			fail(ERR_INVALIDFIRM);
		}

		printf("\n **********************************************************\n");
		printf(" * Warning: about to erase and rewrite your firmware!!!\n");
		printf(" * If this fails for any reason (power goes out, invalid file)\n");
		printf(" * your 64drive will be rendered INOPERABLE!\n");
		printf(" * This must be the first operation performed right after\n");
		printf(" * plugging in USB.\n");
		printf(" * Continue? (y/n) ");
		if(getchar() != 'y') fail(MODE_FPGACONFIG);

		printf("\n - Bulk erasing config (9 seconds)\n");
		memset(tx_buffer, 0, 32);
		tx_buffer[0] = 0x80;	// operation
		tx_buffer[1] = 4;		// length of packet
		tx_buffer[2] = 0xBE;	// magic!
		tx_buffer[3] = 0xEF;
		fail(FT_Write(handle, tx_buffer, 4, &bytes_written));
		sleep(9);
	}
	
	/////////////////////////////////////////////////////
	// write config
	//
	if(mode & MODE_FPGACONFIG){
		int bytes_left = file_size;
		int bytes_done = 0;
		int bytes_do;

		printf(" - Writing FPGA config\n");
		while(1){
			//printf("sdram addr %lu\n", sdram_addr);
			if(bytes_left >= 1024) 
				bytes_do = 1024;
			else
				bytes_do = bytes_left;
			if(bytes_do <= 0) break;
			
			write_header(tx_buffer, 0x70, sdram_addr, bytes_do);
			fail(FT_Write(handle, tx_buffer, 10, &bytes_written));

			fread(buffer, bytes_do, 1, fp);
			fail(FT_Write(handle, buffer, bytes_do, &bytes_written));

			// check for a timeout
			if(bytes_written == 0){
				printf("\n *** Timed out\n");
				printf(" *** Please cycle power and try again\n");
				fail(ERR_TIMEOUT);
			}

			bytes_left -= bytes_written;
			bytes_done += bytes_written;
			sdram_addr += bytes_written;

			// progress bar
			int blocks_done = ((float)bytes_done / file_size)*64;
			int blocks_left = 64-blocks_done;
			printf("\r   ");
			for(int i = 0; i < blocks_done; i++) printf("\xdb");
			for(int i = 0; i < blocks_left; i++) printf("\xb0");
		}
		printf("\n - Success!\n");
		printf(" - Please disconnect the USB cable and plug it back in.\n");
	}
#endif
	
	

	/////////////////////////////////////////////////////
	// start transfer
	//
	if(mode & MODE_LOADIMAGE){
		int bytes_left = file_size;
		int bytes_done = 0;
		int bytes_do;

#if 0
		LARGE_INTEGER liStart;
		LARGE_INTEGER liStop;
		LARGE_INTEGER Frequency;
		QueryPerformanceFrequency(&Frequency);
		QueryPerformanceCounter(&liStart);
#endif

		printf(" - Writing to SDRAM at 0x%lX \n", sdram_addr);
		while(1){
			if(bytes_left >= CHUNK_SIZE) 
				bytes_do = CHUNK_SIZE;
			else
				bytes_do = bytes_left;
			if(bytes_do <= 0) break;
			
			write_header(tx_buffer, 0x10, sdram_addr/2, bytes_do/2);
            status = ftdi_write_data(ftdi, tx_buffer, 10);
			if(status > 0){
				//printf("    Wrote [header] %d bytes ok\n", bytes_written);
                bytes_written = status;
			}else{
				fail(MODE_LOADIMAGE);
			}

			fread(buffer, bytes_do, 1, fp);
            status = ftdi_write_data( ftdi, buffer, bytes_do);
			if(status > 0){
				//printf("    Wrote [data]   %d bytes ok\n", bytes_written);
                bytes_written = status;
			}else{
				fail(MODE_LOADIMAGE);
			}

			// check for a timeout
			if(bytes_written == 0){
				printf("\n *** Timed out\n");
				printf(" *** Please cycle power and try again\n");
				fail(ERR_TIMEOUT);
			}

			bytes_left -= bytes_written;
			bytes_done += bytes_written;
			sdram_addr += bytes_written;

			// progress bar
			int blocks_done = ((float)bytes_done / file_size)*64;
			int blocks_left = 64-blocks_done;
			//printf("\r   ");
			//for(int i = 0; i < blocks_done; i++) printf("\xdb");
			//for(int i = 0; i < blocks_left; i++) printf("\xb0");
		}

#if 0
		// stop the timer
		QueryPerformanceCounter(&liStop);
		long long llTimeDiff = liStop.QuadPart - liStart.QuadPart;
		// get the difference of the timer
		double dftDuration = (double) llTimeDiff * 1000.0 / (double) Frequency.QuadPart;
		// erase progress bar
		printf("\r");	for(int i = 0; i < 79; i++) printf(" "); printf("\r");
		printf(" - Completed in %.2f seconds (%.2f MB/sec)\n", dftDuration/1000, file_size/1024/1024/(dftDuration/1000));
#endif
		//printf("\r");	for(int i = 0; i < 79; i++) printf(" "); printf("\r");
        printf(" - Completed\n");
		
	}
	printf("\n - Done\n");

	////////////////////////////////////////////////////
	// cleanup
	//
	fclose(fp);
    ftdi_list_free(&devInfo);
    ftdi_free(ftdi);
    ftdi = NULL;


	return 0;
}

void fail(int st)
{ 
	if(st == 0) return;

	printf("\n *** \n ***  FAIL: %d\n ***\n", st);
	exit(-1);
}

